﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;

namespace imaeg {
    class Interpreter {

        public static Bitmap load(string path) {
            if (path.ToLower().EndsWith(".tga")) return Targa(path);
            return new Bitmap(path);
        }

        // Tested on files created by:
        // - photoshop
        // - mspaint
        // - TF2
        public static Bitmap Targa(string path) {
            byte[] src = System.IO.File.ReadAllBytes(path);
            int w = src[0x0D] * 256 + src[0x0C];
            int h = src[0x0F] * 256 + src[0x0E];
            bool flip = src[0x11] == 0x20;
            Bitmap bm = new Bitmap(w, h);
            BitmapData bd = bm.LockBits(new Rectangle(0, 0, w, h),
                System.Drawing.Imaging.ImageLockMode.WriteOnly,
                System.Drawing.Imaging.PixelFormat.Format24bppRgb);
            int sp;
            unsafe {
                for (int y = 0; y < h; y++) {
                    if (flip) sp = 0x12 + 3 * w * y;
                    else sp = 0x12 + 3 * w * (h - y - 1);
                    byte* dst = (byte*)bd.Scan0 + (bd.Stride * y);
                    for (int n = 0; n < 3 * w; n++) {
                        dst[n] = src[n + sp];
                    }
                }
            }
            bm.UnlockBits(bd);
            return bm;
        }

    }
}
